
/* Chrysalide - Outil d'analyse de fichiers binaires
 * symbol.h - prototypes pour la gestion des symboles dans un binaire
 *
 * Copyright (C) 2009-2017 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  Chrysalide is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Chrysalide is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */


#ifndef _FORMAT_SYMBOL_H
#define _FORMAT_SYMBOL_H


#include <glib-object.h>


#include "../analysis/routine.h"
#include "../analysis/db/item.h"
#include "../analysis/db/items/comment.h"
#include "../arch/instruction.h"
#include "../glibext/linegen.h"



/* --------------------- FONCTIONNALITES BASIQUES POUR SYMBOLES --------------------- */


/* Types de symbole */
typedef enum _SymbolType
{
    STP_DATA,                               /* Données brutes              */
    STP_ROUTINE,                            /* Simple morceau de code      */
    STP_CODE_LABEL,                         /* Renvoi au sein de code      */
    STP_OBJECT,                             /* Objet quelconque            */
    STP_ENTRY_POINT,                        /* Morceau de code en entrée   */
    STP_STRING,                             /* Chaîne de caractères        */
    STP_RO_STRING,                          /* Chaîne de caractères        */

    STP_COUNT

} SymbolType;


#define HAS_DATA_INSTR(type) (type == STP_DATA || type == STP_RO_STRING)



#define G_TYPE_BIN_SYMBOL               g_binary_symbol_get_type()
#define G_BIN_SYMBOL(obj)               (G_TYPE_CHECK_INSTANCE_CAST((obj), g_binary_symbol_get_type(), GBinSymbol))
#define G_IS_BIN_SYMBOL(obj)            (G_TYPE_CHECK_INSTANCE_TYPE((obj), g_binary_symbol_get_type()))
#define G_BIN_SYMBOL_CLASS(klass)       (G_TYPE_CHECK_CLASS_CAST((klass), G_TYPE_BIN_SYMBOL, GBinSymbolClass))
#define G_IS_BIN_SYMBOL_CLASS(klass)    (G_TYPE_CHECK_CLASS_TYPE((klass), G_TYPE_BIN_SYMBOL))
#define G_BIN_SYMBOL_GET_CLASS(obj)     (G_TYPE_INSTANCE_GET_CLASS((obj), G_TYPE_BIN_SYMBOL, GBinSymbolClass))


/* Symbole d'exécutable (instance) */
typedef struct _GBinSymbol GBinSymbol;

/* Symbole d'exécutable (classe) */
typedef struct _GBinSymbolClass GBinSymbolClass;


/* Indique le type défini pour un symbole d'exécutable. */
GType g_binary_symbol_get_type(void);

/* Crée un nouveau symbole d'exécutable. */
GBinSymbol *g_binary_symbol_new(SymbolType);

/* Compare deux symboles d'exécutable selon leurs propriétés. */
int g_binary_symbol_cmp(const GBinSymbol * const *, const GBinSymbol * const *);

/* Compare un symbole et une localisation. */
int g_binary_symbol_cmp_with_vmpa(const GBinSymbol *, const vmpa2t *);

/* Fournit le type du symbole. */
SymbolType g_binary_symbol_get_target_type(const GBinSymbol *);

/* Fournit un étiquette pour viser un symbole. */
const char *g_binary_symbol_get_label(const GBinSymbol *);

/* Définit un autre nom pour le symbole. */
void g_binary_symbol_set_alt_label(GBinSymbol *, const char *);

/* Fournit l'emplacement où se situe un symbole. */
const mrange_t *g_binary_symbol_get_range(const GBinSymbol *);

/* Attache la routine associée au symbole. */
void _g_binary_symbol_attach_routine(GBinSymbol *, GBinRoutine *, SymbolType);

/* Attache la routine associée au symbole. */
void g_binary_symbol_attach_routine(GBinSymbol *, GBinRoutine *);

/* Attache l'instruction associée au symbole. */
void g_binary_symbol_attach_instruction(GBinSymbol *, GArchInstruction *);

/* Fournit l'éventuelle routine associée au symbole. */
GBinRoutine *g_binary_symbol_get_routine(const GBinSymbol *);

/* Fournit l'éventuelle instruction associée au symbole. */
GArchInstruction *g_binary_symbol_get_instruction(const GBinSymbol *);


/**
 * Confort pour l'ajout de symboles basés sur des formats.
 */




#define ADD_STR_AS_SYM(_fmt, _sym, _ins)                                    \
    ({                                                                      \
        bool __result;                                                      \
        _sym = g_binary_symbol_new(STP_RO_STRING);                          \
        g_binary_symbol_attach_instruction(_sym, _ins);                     \
        __result = g_binary_format_add_symbol(G_BIN_FORMAT(_fmt), _sym);    \
        __result;                                                           \
    })



/* ------------------------ OFFRE DE CAPACITES DE GENERATION ------------------------ */


/* Détermine si un symbole pour faire office de générateur. */
GLineGenerator *g_binary_symbol_produce_label(GBinSymbol *);



#endif  /* _FORMAT_SYMBOL_H */
